#     Copyright (c) <2021> <University of Paderborn>
#     Signal and System Theory Group, Univ. of Paderborn, https://sst-group.org/
#     https://github.com/SSTGroup/independent_vector_analysis
#
#     Permission is hereby granted, free of charge, to any person
#     obtaining a copy of this software and associated documentation
#     files (the "Software"), to deal in the Software without restriction,
#     including without limitation the rights to use, copy, modify and
#     merge the Software, subject to the following conditions:
#
#     1.) The Software is used for non-commercial research and
#        education purposes.
#
#     2.) The above copyright notice and this permission notice shall be
#        included in all copies or substantial portions of the Software.
#
#     3.) Publication, Distribution, Sublicensing, and/or Selling of
#        copies or parts of the Software requires special agreements
#        with the University of Paderborn and is in general not permitted.
#
#     4.) Modifications or contributions to the software must be
#        published under this license. The University of Paderborn
#        is granted the non-exclusive right to publish modifications
#        or contributions in future versions of the Software free of charge.
#
#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#     EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#     OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#     NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#     HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#     WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#     FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#     OTHER DEALINGS IN THE SOFTWARE.
#
#     Persons using the Software are encouraged to notify the
#     Signal and System Theory Group at the University of Paderborn
#     about bugs. Please reference the Software in your publications
#     if it was used for them.


import numpy as np
import scipy as sc
import time

from .helpers_iva import _normalize_column_vectors, _decouple_trick, _bss_isi, whiten_data, \
    _resort_scvs
from .initializations import _jbss_sos, _cca


def iva_g(X, opt_approach='newton', complex_valued=False, circular=False, whiten=True,
          verbose=False, A=None, W_init=None, jdiag_initW=False, max_iter=1024,
          W_diff_stop=1e-6, alpha0=1.0, return_W_change=False):
    """
    Implementation of all the second-order (Gaussian) independent vector analysis (IVA) algorithms.
    Namely real-valued and complex-valued with circular and non-circular using Newton, gradient,
    and quasi-Newton optimizations.
    For a general description of the algorithm and its relationship with others,
    see http://mlsp.umbc.edu/jointBSS_introduction.html.


    Parameters
    ----------
    X : np.ndarray
        data matrix of dimensions N x T x K.
        Data observations are from K data sets, i.e., X[k] = A[k] @ S[k], where A[k] is an N x N
        unknown invertible mixing matrix, and S[k] is N x T matrix with the nth row corresponding to
        T samples of the nth source in the kth dataset. This enforces the assumption of an equal
        number of samples in each dataset.
        For IVA, it is assumed that a source is statistically independent of all the other sources
        within the dataset and exactly dependent on at most one source in each of the other
        datasets.

    opt_approach : str, optional
        optimization type: 'gradient','newton','quasi'

    complex_valued : bool, optional
        if True, data pseudo cross-covariance matrices are calculated.
        If any input is complex, then complex_valued is forced to True

    circular : bool, optional
        set to True to only consider circular for complex-valued cost function

    whiten : bool, optional
        if True, data is whitened.
        For the complex-valued gradient and for quasi approach, whiten is forced to True

    verbose : bool, optional
        enables print statements if True. Automatically set to True if A is provided

    A : np.ndarray, optional
        true mixing matrices A of dimensions N x N x K, automatically sets verbose to True

    W_init : np.ndarray, optional
        initial estimate for demixing matrices in W, with dimensions N x N x K

    jdiag_initW : bool, optional
        if True, use CCA (K=2) / joint diagonalization (K>2) for initialization, else random

    max_iter : int, optional
        max number of iterations

    W_diff_stop : float, optional
        stopping criterion

    alpha0 : float, optional
        initial step size scaling (will be doubled for complex-valued gradient)

    return_W_change : bool, optional
        if the change in W in each iteration should be returned


    Returns
    -------
    W : np.ndarray
        the estimated demixing matrices of dimensions N x N x K so that ideally
        W[k] @ A[k] = P @ D[k], where P is any arbitrary permutation matrix and D[k] is any
        diagonal invertible (scaling) matrix. Note that P is common to all datasets. This is
        to indicate that the local permutation ambiguity between dependent sources
        across datasets should ideally be resolved by IVA.

    cost : float
        the cost for each iteration

    Sigma_N : np.ndarray
        Covariance matrix of each source component vector, with dimensions K x K x N

    isi : float
        joint isi (inter-symbol-interference) for each iteration. Only available if user
        supplies true mixing matrices for computing a performance metric. Else returns np.nan.


    Notes
    -----

    Coded by Matthew Anderson (matt.anderson@umbc.edu)
    Converted to Python by Isabell Lehmann (isabell.lehmann@sst.upb.de)

    References:

    [1] M. Anderson, X.-L. Li, & T. Adalı, "Nonorthogonal Independent Vector Analysis Using
    Multivariate Gaussian Model," LNCS: Independent Component Analysis and Blind Signal Separation,
    Latent Variable Analysis and Signal Separation, Springer Berlin / Heidelberg, 2010, 6365,
    354-361
    [2] M. Anderson, T. Adalı, & X.-L. Li, "Joint Blind Source Separation of Multivariate Gaussian
    Sources: Algorithms and Performance Analysis," IEEE Trans. Signal Process., 2012, 60, 1672-1683
    [3] M. Anderson, X.-L. Li, & T. Adalı, "Complex-valued Independent Vector Analysis: Application
    to Multivariate Gaussian Model," Signal Process., 2012, 1821-1831

    Version 01 - 20120913 - Initial publication
    Version 02 - 20120919 - Using decouple_trick function, bss_isi, whiten_data, & cca subfunctions
    Version 03 - 20210129 - Python version of the code, with max_iter=1024.

    """

    start = time.time()

    if X.ndim != 3:
        raise AssertionError('X must have dimensions N x T x K.')
    elif X.shape[2] == 1:
        raise AssertionError('There must be ast least K=2 datasets.')

    if opt_approach != 'newton' and opt_approach != 'gradient' and opt_approach != 'quasi':
        raise AssertionError("opt_approach must be 'newton', 'gradient' or 'quasi'")

    # get dimensions
    N, T, K = X.shape

    if A is not None:
        supply_A = True
        if A.shape[0] != N or A.shape[1] != N or A.shape[2] != K:
            raise AssertionError('A must have dimensions N x N x K.')
    else:
        supply_A = False

    blowup = 1e3
    # set alpha0 to max(alpha_min, alpha0*alpha_scale) when cost does not decrease
    alpha_scale = 0.9
    alpha_min = W_diff_stop

    # if any input is complex, then 'complex_valued' must be True
    complex_valued = complex_valued or np.any(np.iscomplex(A)) or np.any(np.iscomplex(X))

    # whitening is required for the quasi & complex-valued gradient approach
    whiten = whiten or (opt_approach == 'quasi') or (
            complex_valued and opt_approach == 'gradient')

    if whiten:
        X, V = whiten_data(X)

    # calculate cross-covariance matrices of X
    R_xx = np.zeros((N, N, K, K), dtype=X.dtype)
    for k1 in range(K):
        for k2 in range(k1, K):
            R_xx[:, :, k1, k2] = 1 / T * X[:, :, k1] @ np.conj(X[:, :, k2].T)
            R_xx[:, :, k2, k1] = np.conj(R_xx[:, :, k1, k2].T)  # R_xx is Hermitian

    # Check rank of data-covariance matrix: should be full rank, if not we inflate (this is ad hoc)
    # concatenate all covariance matrices in a big matrix
    R_xx_all = np.moveaxis(R_xx, [0, 1, 2, 3], [0, 2, 1, 3]).reshape(
        (R_xx.shape[0] * R_xx.shape[2], R_xx.shape[1] * R_xx.shape[3]), order='F')
    rank = np.linalg.matrix_rank(R_xx_all)
    if rank < (N * K):
        # inflate Rx
        _, k, _ = np.linalg.svd(R_xx_all)
        R_xx_all += k[rank - 1] * np.eye(N * K)  # add smallest singular value to main diagonal
        R_xx = np.moveaxis(
            R_xx_all.reshape(R_xx.shape[0], R_xx.shape[2], R_xx.shape[1], R_xx.shape[3], order='F'),
            [0, 2, 1, 3], [0, 1, 2, 3])

    # complex-valued non-circular cost
    if complex_valued and not circular:
        # compute data pseudo cross-covariance matrices
        P_xx = np.zeros((N, N, K, K), dtype=complex)
        for k1 in range(K):
            for k2 in range(k1, K):
                P_xx[:, :, k1, k2] = 1 / T * X[:, :, k1] @ X[:, :, k2].T
                P_xx[:, :, k2, k1] = P_xx[:, :, k1, k2].T  # P_xx is symmetric

    # Initializations

    # Initialize W
    if W_init is not None:
        if W_init.shape[0] != N or W_init.shape[1] != N or (
                W_init.ndim == 3 and W_init.shape[2] != K):
            raise AssertionError('W_init must have dimension N x N or N x N x K.')

        W = np.copy(W_init)

        if not complex_valued and np.any(np.iscomplex(W)) and opt_approach == 'gradient':
            raise Warning('Supplied initial W is complex-valued and gradient option_approach is '
                          'set. Whitening option should be set to True for best performance.')
        if W.shape[0] == N and W.shape[1] == N and W.ndim == 2:
            W = np.tile(W[:, :, np.newaxis], [1, 1, K])

        if whiten:
            for k in range(K):
                W[:, :, k] = np.linalg.solve(V[:, :, k].T, W[:, :, k].T).T

    else:
        if jdiag_initW:
            if K > 2:
                # initialize with multi-set diagonalization (orthogonal solution)
                W = _jbss_sos(X, 0, 'whole')

            else:
                W = _cca(X)

        else:
            # randomly initialize
            W = np.random.randn(N, N, K)
            if complex_valued:
                W = W + 1j * np.random.randn(N, N, K)
            for k in range(K):
                W[:, :, k] = np.linalg.solve(sc.linalg.sqrtm(W[:, :, k] @ W[:, :, k].T), W[:, :, k])

    if supply_A:
        # only reason to supply A matrices is to display running performance and to compute
        # output_isi
        verbose = True
        isi = np.zeros(max_iter)
        if whiten:
            # A matrix is conditioned by V if data is whitened
            A_w = np.copy(A)
            for k in range(K):
                A_w[:, :, k] = V[:, :, k] @ A_w[:, :, k]
        else:
            A_w = np.copy(A)
    else:
        isi = np.nan

    # Initialize some local variables
    cost = np.zeros(max_iter)
    cost_const = K * np.log(2 * np.pi * np.exp(1))  # local constant

    # Initializations based on real vs complex-valued
    if complex_valued:
        grad = np.zeros((N, K), dtype=complex)
        if opt_approach == 'newton':
            HA = np.zeros((N * K, N * K), dtype=complex)
        elif opt_approach == 'gradient':
            # Double step-size for complex-valued gradient optimization
            alpha0 = 2 * alpha0
        H = np.zeros((N * K, N * K), dtype=complex)
    else:  # real-valued
        grad = np.zeros((N, K))
        if opt_approach == 'newton':
            H = np.zeros((N * K, N * K))

    # to store the change in W in each iteration
    W_change = []

    # Main Iteration Loop
    for iteration in range(max_iter):
        term_criterion = 0

        # Some additional computations of performance via ISI when true A is supplied
        if supply_A:
            avg_isi, joint_isi = _bss_isi(W, A_w)
            isi[iteration] = joint_isi

        W_old = np.copy(W)  # save current W as W_old
        cost[iteration] = 0
        for k in range(K):
            cost[iteration] -= np.log(np.abs(np.linalg.det(W[:, :, k])))

        Q = 0
        R = 0

        # Loop over each SCV
        for n in range(N):
            Wn = np.conj(W[n, :, :]).flatten(order='F')

            # Efficient version of Sigma_n = 1/T * Y_n @ np.conj(Y_n.T) with Y_n = W_n @ X_n
            if complex_valued:
                Sigma_n = np.eye(K, dtype=complex)
            else:
                Sigma_n = np.eye(K)

            for k1 in range(K):
                for k2 in range(k1, K):
                    Sigma_n[k1, k2] = W[n, :, k1] @ R_xx[:, :, k1, k2] @ np.conj(W[n, :, k2])
                    Sigma_n[k2, k1] = np.conj(Sigma_n[k1, k2])

            if complex_valued and not circular:
                # complex-valued non-circular cost
                # compute SCV pseudo cross-covariance matrices
                Sigma_P_n = np.zeros((K, K), dtype=complex)  # pseudo = 1/T * Y_n @ Y_n.T
                for k1 in range(K):
                    for k2 in range(k1, K):
                        Sigma_P_n[k1, k2] = W[n, :, k1] @ P_xx[:, :, k1, k2] @ W[n, :, k2]
                        Sigma_P_n[k2, k1] = Sigma_P_n[k1, k2]

                Sigma_total = np.vstack(
                    [np.hstack([Sigma_n, Sigma_P_n]),
                     np.hstack([np.conj(Sigma_P_n.T), np.conj(Sigma_n)])])
                cost[iteration] += 0.5 * np.log(np.abs(np.linalg.det(Sigma_total)))
            else:
                cost[iteration] += 0.5 * (cost_const + np.log(np.abs(np.linalg.det(Sigma_n))))

            if complex_valued and not circular:
                Sigma_inv = np.linalg.inv(Sigma_total)
                P = Sigma_inv[0:K, 0:K]
                Pt = Sigma_inv[0:K, K:2 * K]
            else:
                Sigma_inv = np.linalg.inv(Sigma_n)

            hnk, Q, R = _decouple_trick(W, n, Q, R)

            # Loop over each dataset
            for k in range(K):
                # Analytic derivative of cost function with respect to vn
                # Code below is efficient implementation of computing the gradient, which is
                # independent of T
                grad[:, k] = - hnk[:, k] / (W[n, :, k] @ hnk[:, k])

                if complex_valued and not circular:
                    for kk in range(K):
                        grad[:, k] += R_xx[:, :, k, kk] @ np.conj(W[n, :, kk]) * np.conj(P[k, kk]) \
                                      + P_xx[:, :, k, kk] @ W[n, :, kk] * np.conj(Pt[k, kk])
                else:
                    for kk in range(K):
                        grad[:, k] += R_xx[:, :, k, kk] @ np.conj(W[n, :, kk]) * Sigma_inv[kk, k]

                if opt_approach == 'gradient' or (opt_approach == 'quasi' and not complex_valued):
                    wnk = np.conj(W[n, :, k])
                    if opt_approach == 'gradient':
                        grad_norm = _normalize_column_vectors(grad[:, k])
                        grad_norm_proj = _normalize_column_vectors(grad_norm - np.conj(
                            wnk) @ grad_norm * wnk)  # non-colinear direction normalized
                        W[n, :, k] = np.conj(
                            _normalize_column_vectors(wnk - alpha0 * grad_norm_proj))

                        for kk in range(K):  # = 1/T * Y_n @ np.conj(Yn.T)
                            Sigma_n[k, kk] = W[n, :, k] @ R_xx[:, :, k, kk] @ np.conj(W[n, :, kk].T)
                        Sigma_n[:, k] = np.conj(Sigma_n[k, :].T)

                        if complex_valued and not circular:
                            for kk in range(K):  # = 1/T * Y_n @ * Y_n.T
                                Sigma_P_n[k, kk] = W[n, :, k] @ P_xx[:, :, k, kk] @ W[n, :, kk].T
                            Sigma_P_n[:, k] = Sigma_P_n[k, :].T
                            Sigma_total = np.vstack(
                                [np.hstack([Sigma_n, Sigma_P_n]),
                                 np.hstack([np.conj(Sigma_P_n.T), np.conj(Sigma_n)])])
                            Sigma_inv = np.linalg.inv(Sigma_total)
                        else:
                            Sigma_inv = np.linalg.inv(Sigma_n)

                    else:  # real-valued quasi (block newton)
                        # Hessian inverse computed using matrix inversion lemma
                        H_inv = 1 / Sigma_inv[k, k] * (
                                np.eye(N) - np.outer(hnk[:, k], hnk[:, k]) / (
                                Sigma_inv[k, k] + 1 / (hnk[:, k] @ wnk) ** 2))

                        # Block-Newton update of Wnk
                        W[n, :, k] = _normalize_column_vectors(wnk - alpha0 * (H_inv @ grad[:, k]))

            if opt_approach == 'newton' or (opt_approach == 'quasi' and complex_valued):
                # Compute SCV Hessian
                for k1 in range(K):
                    if complex_valued:
                        if opt_approach == 'newton':  # complex-valued Newton
                            HA[k1 * N:k1 * N + N,
                            k1 * N:k1 * N + N] = np.outer(
                                np.conj(hnk[:, k1]), np.conj(hnk[:, k1])) / (
                                                     np.conj(hnk[:, k1] @ W[n, :, k1])) ** 2

                            if not circular:  # complex-valued Newton non-circular
                                HA[k1 * N:k1 * N + N, k1 * N:k1 * N + N] += Pt[k1, k1] * np.conj(
                                    P_xx[:, :, k1, k1])
                        H[k1 * N:(k1 + 1) * N, k1 * N:(k1 + 1) * N] = np.conj(
                            Sigma_inv[k1, k1] * R_xx[:, :, k1, k1])
                    else:  # real-valued Newton
                        H[k1 * N:k1 * N + N, k1 * N:k1 * N + N] = \
                            Sigma_inv[k1, k1] * R_xx[:, :, k1, k1] + np.outer(
                                hnk[:, k1], hnk[:, k1]) / (hnk[:, k1] @ W[n, :, k1]) ** 2

                    for k2 in range(k1 + 1, K):
                        if opt_approach == 'newton' and complex_valued and not circular:
                            # complex-valued Newton non-circular
                            HA[k1 * N: k1 * N + N, k2 * N: k2 * N + N] = Pt[k2, k1] * np.conj(
                                P_xx[:, :, k1, k2])
                            HA[k2 * N: k2 * N + N, k1 * N: k1 * N + N] = Pt[k1, k2] * np.conj(
                                P_xx[:, :, k2, k1])
                            Hs = np.conj(P[k2, k1] * R_xx[:, :, k1, k2])
                        else:
                            Hs = Sigma_inv[k1, k2] * R_xx[:, :, k2, k1].T
                        H[k1 * N: k1 * N + N, k2 * N: k2 * N + N] = Hs
                        H[k2 * N: k2 * N + N, k1 * N: k1 * N + N] = np.conj(Hs.T)

                # Newton Update
                if not complex_valued:  # real-valued Newton
                    Wn -= alpha0 * np.linalg.solve(H, grad.flatten('F'))
                else:  # complex-valued
                    if opt_approach == 'newton':  # complex-valued Newton
                        # improve stability by using 0.99 in Hessian
                        Wn -= alpha0 * np.linalg.solve(
                            np.conj(H) - 0.99 * np.conj(HA) @ np.linalg.solve(H, HA),
                            grad.flatten('F') - np.conj(HA) @ np.linalg.solve(H, np.conj(
                                grad.flatten('F'))))
                    elif opt_approach == 'quasi':  # complex-valued Quasi-Newton
                        Wn -= alpha0 * np.linalg.solve(np.conj(H), grad.flatten('F'))
                    else:
                        raise AssertionError('Should not happen.')
                    Wn = np.conj(Wn)

                # Store Updated W
                Wn = np.reshape(Wn, (N, K), 'F')
                for k in range(K):
                    W[n, :, k] = _normalize_column_vectors(Wn[:, k])

        for k in range(K):
            if complex_valued:  # for complex data, W_old @ W.T is not bounded between -1 and 1
                term_criterion = np.maximum(term_criterion, np.amax(np.abs(
                    1 - np.abs(np.diag(W_old[:, :, k] @ np.conj(W[:, :, k].T))))))
            else:  # in original matlab code, this is the term criterion
                # (the result is the same as above for real data)
                term_criterion = np.maximum(term_criterion, np.amax(
                    1 - np.abs(np.diag(W_old[:, :, k] @ np.conj(W[:, :, k].T)))))

        W_change.append(term_criterion)

        # Decrease step size alpha if cost increased from last iteration
        if iteration > 0 and cost[iteration] > cost[iteration - 1]:
            alpha0 = np.maximum(alpha_min, alpha_scale * alpha0)

        # Check the termination condition
        if term_criterion < W_diff_stop:
            break
        elif term_criterion > blowup or np.isnan(cost[iteration]):
            for k in range(K):
                W[:, :, k] = np.eye(N) + 0.1 * np.random.randn(N, N)
            if complex_valued:
                W = W + 0.1j * np.random.randn(N, N, K)
            if verbose:
                print('W blowup, restart with new initial value.')

        # Display Iteration Information
        if verbose:
            if supply_A:
                print(
                    f'Step {iteration}: W change: {term_criterion}, Cost: {cost[iteration]}, '
                    f'Avg ISI: {avg_isi}, Joint ISI: {joint_isi}')
            else:
                print(f'Step {iteration}: W change: {term_criterion}, Cost: {cost[iteration]}')

    # Finish Display
    if iteration == 0 and verbose:
        if supply_A:
            print(
                f'Step {iteration}: W change: {term_criterion}, Cost: {cost[iteration]}, '
                f'Avg ISI: {avg_isi}, Joint ISI: {joint_isi}')
        else:
            print(f'Step {iteration}: W change: {term_criterion}, Cost: {cost[iteration]}')

    # Clean-up Outputs
    cost = cost[0:iteration + 1]

    if supply_A:
        isi = isi[0:iteration + 1]

    if whiten:
        for k in range(K):
            W[:, :, k] = W[:, :, k] @ V[:, :, k]
    else:  # no prewhitening
        # Scale demixing vectors to generate unit variance sources
        for n in range(N):
            for k in range(K):
                W[n, :, k] /= np.sqrt(W[n, :, k] @ R_xx[:, :, k, k] @ np.conj(W[n, :, k]))

    # Resort order of SCVs: Order the components from most to least ill-conditioned
    if not complex_valued or circular:
        P_xx = None
    if not whiten:
        V = None
    W, Sigma_N = _resort_scvs(W, R_xx, whiten, V, complex_valued, circular, P_xx)

    end = time.time()
    print(f"IVA-G finished after {(end - start) / 60} minutes with {iteration} iterations.")

    if return_W_change:
        return W, cost, Sigma_N, isi, W_change
    else:
        return W, cost, Sigma_N, isi



#     Copyright (c) <2021> <University of Paderborn>
#     Signal and System Theory Group, Univ. of Paderborn, https://sst-group.org/
#     https://github.com/SSTGroup/independent_vector_analysis
#
#     Permission is hereby granted, free of charge, to any person
#     obtaining a copy of this software and associated documentation
#     files (the "Software"), to deal in the Software without restriction,
#     including without limitation the rights to use, copy, modify and
#     merge the Software, subject to the following conditions:
#
#     1.) The Software is used for non-commercial research and
#        education purposes.
#
#     2.) The above copyright notice and this permission notice shall be
#        included in all copies or substantial portions of the Software.
#
#     3.) Publication, Distribution, Sublicensing, and/or Selling of
#        copies or parts of the Software requires special agreements
#        with the University of Paderborn and is in general not permitted.
#
#     4.) Modifications or contributions to the software must be
#        published under this license. The University of Paderborn
#        is granted the non-exclusive right to publish modifications
#        or contributions in future versions of the Software free of charge.
#
#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#     EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#     OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#     NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#     HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#     WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#     FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#     OTHER DEALINGS IN THE SOFTWARE.
#
#     Persons using the Software are encouraged to notify the
#     Signal and System Theory Group at the University of Paderborn
#     about bugs. Please reference the Software in your publications
#     if it was used for them.


import numpy as np
import scipy as sc
import time

from .helpers_iva import _normalize_column_vectors, whiten_data, _bss_isi, _decouple_trick, \
    _comp_l_sos_cost, _resort_scvs
from .initializations import _jbss_sos, _cca


def iva_l_sos(X, whiten=True, grad_projection=False, verbose=False, A=None, W_init=None,
              jdiag_initW=False, max_iter=2 * 512, termination_criterion='change_in_W',
              W_diff_stop=1e-4, alpha0=1.0, return_W_change=False):
    """
    Implementation of the independent vector analysis (IVA) algorithm using a correlated
    multivariate Laplace prior. This implementation also makes use of decoupling method to
    achieve 'Newton' like convergence (see references below).
    This implementation only works for real-valued data.
    For a general description of the algorithm and its relationship with others,
    see http://mlsp.umbc.edu/jointBSS_introduction.html.

    Parameters
    ----------
    X : np.ndarray
         data matrix of dimensions N x T x K.
         Real-valued data observations are from K data sets, i.e., X[k] = A[k] @ S[k], where A[k]
         is an N x N unknown invertible mixing matrix, and S[k] is N x T matrix with the nth row
         corresponding to T samples of the nth source in the kth dataset. This enforces the
         assumption of an equal number of samples in each dataset.
         For IVA, it is assumed that a source is statistically independent of all the other sources
         within the dataset and exactly dependent on at most one source in each of the other
         datasets.

    whiten : bool, optional
         if True, data is whitened. Whitening significantly decreases the runtime of the algorithm.

    grad_projection : bool, optional
        if True, the non-colinear direction is normalized

    verbose : bool, optional
         enables print statements if True. Automatically set to True if A is provided

    A : np.ndarray, optional
         true mixing matrices A of dimensions N x N x K, automatically sets verbose to True

    W_init : np.ndarray, optional
         initial estimate for demixing matrices in W, with dimensions N x N x K

    jdiag_initW : bool, optional
        if True, use CCA (K=2) / joint diagonalization (K>2) for initialization, else random

    max_iter : int, optional
         max number of iterations

    termination_criterion: str, optional
       if termination depends on W ('change_in_W') or on cost ('change_in_cost')

    W_diff_stop : float, optional
        termination threshold

    alpha0 : float, optional
         initial step size scaling (will be doubled for complex-valued gradient)

    return_W_change : bool, optional
        if the change in W in each iteration should be returned


    Returns
    -------
     W : np.ndarray
         the estimated demixing matrices of dimensions N x N x K so that ideally
         W[k] @ A[k] = P @ D[k], where P is any arbitrary permutation matrix and D[k] is any
         diagonal invertible (scaling) matrix. Note that P is common to all datasets. This is
         to indicate that the local permutation ambiguity between dependent sources
         across datasets should ideally be resolved by IVA.

     cost : float
         the cost for each iteration

     Sigma_N : np.ndarray
        Covariance matrix of each source component vector, with dimensions K x K x N

     isi : float
         joint inter-symbol-interference for each iteration. Only available if user supplies true
         mixing matrices for computing a performance metric. Else returns np.nan.


    Notes
    -----

    Coded by Matthew Anderson (matt.anderson@umbc.edu)
    Modified by Suchita Bhinge (suchita1@umbc.edu) on September 14 2018:
    Modification is in the score function in order to account for effect on
    large number of datasets
    Converted to Python by Isabell Lehmann (isabell.lehmann@sst.upb.de)

    References:

    [1] S. Bhinge, R. Mowakeaa, V. D. Calhoun, & T. Adali, "Extraction of time-varying
    spatio-temporal networks using parameter-tuned constrained IVA," Transaction on Medical Imaging,
    vol. 38, pp: 1715-1725, 2019
    [2] M. Anderson, T. Adali, & X.-L. Li,  "Joint Blind Source Separation of Multivariate Gaussian
    Sources: Algorithms and Performance Analysis," IEEE Trans. Signal Process., 2012, 60, 1672-1683
    [3] M. Anderson, G.-S. Fu, R. Phlypo, and T. Adali, "Independent vector analysis: Identification
    conditions and performance bounds," IEEE Trans. Signal Processing, vol. 62, pp. 4399-4410,
    Sep. 2014.

    Version 01 - 20120919 - Initial publication.
    Version 02 - 20140806 - Improved notation in comments.
    Version 03 - 20210301 - Python version.

    """

    start = time.time()

    if X.ndim != 3:
        raise AssertionError('X must have dimensions N x T x K.')
    elif X.shape[2] == 1:
        raise AssertionError('There must be ast least K=2 datasets.')

    if termination_criterion != 'change_in_W' and termination_criterion != 'change_in_cost':
        raise AssertionError("termination_criterion must be 'change_in_W' or 'change_in_cost'")

    # get dimensions
    N, T, K = X.shape

    if A is not None:
        supply_A = True
        if A.shape[0] != N or A.shape[1] != N or A.shape[2] != K:
            raise AssertionError('A must have dimensions N x N x K.')
    else:
        supply_A = False

    blowup = 1e3
    # set alpha0 to max(alpha_min, alpha0*alpha_scale) when cost does not decrease
    alpha_scale = 0.9
    alpha_min = W_diff_stop

    if whiten:
        X, V = whiten_data(X)

    # calculate cross-covariance matrices of X (of shape N x N x K x K)
    R_xx = 1 / T * np.einsum('NTK, nTk -> NnKk', X, X)

    # Check rank of data-covariance matrix: should be full rank, if not we inflate (this is ad hoc)
    # concatenate all covariance matrices in a big matrix
    R_xx_all = np.moveaxis(R_xx, [0, 1, 2, 3], [0, 2, 1, 3]).reshape(
        (R_xx.shape[0] * R_xx.shape[2], R_xx.shape[1] * R_xx.shape[3]), order='F')
    rank = np.linalg.matrix_rank(R_xx_all)
    if rank < (N * K):
        # inflate Rx
        _, k, _ = np.linalg.svd(R_xx_all)
        R_xx_all += k[rank - 1] * np.eye(N * K)  # add smallest singular value to main diagonal
        R_xx = np.moveaxis(
            R_xx_all.reshape(R_xx.shape[0], R_xx.shape[2], R_xx.shape[1], R_xx.shape[3], order='F'),
            [0, 2, 1, 3], [0, 1, 2, 3])

    # Initializations

    # Initialize W
    if W_init is not None:
        if W_init.shape[0] != N or W_init.shape[1] != N or (
                W_init.ndim == 3 and W_init.shape[2] != K):
            raise AssertionError('W_init must have dimension N x N or N x N x K.')

        W = np.copy(W_init)

        if W.shape[0] == N and W.shape[1] == N and W.ndim == 2:
            W = np.tile(W[:, :, np.newaxis], [1, 1, K])

        if whiten:
            for k in range(K):
                W[:, :, k] = np.linalg.solve(V[:, :, k].T, W[:, :, k].T).T

    else:
        if jdiag_initW:
            if K > 2:
                # initialize with multi-set diagonalization (orthogonal solution)
                W = _jbss_sos(X, 0, 'whole')

            else:
                W = _cca(X)

        else:
            # randomly initialize
            W = np.random.randn(N, N, K)
            if X.dtype == complex:
                W = W + 1j * np.random.randn(N, N, K)
            for k in range(K):
                W[:, :, k] = np.linalg.solve(sc.linalg.sqrtm(W[:, :, k] @ W[:, :, k].T), W[:, :, k])

    if supply_A:
        # only reason to supply A matrices is to display running performance and to compute
        # output_isi
        verbose = True
        isi = np.zeros(max_iter)
        if whiten:
            # A matrix is conditioned by V if data is whitened
            A_w = np.copy(A)
            for k in range(K):
                A_w[:, :, k] = V[:, :, k] @ A_w[:, :, k]
        else:
            A_w = np.copy(A)
    else:
        isi = np.nan

    # Initialize some local variables
    cost = np.zeros(max_iter)
    gammaShape = (K + 1) ** 0.5  # .. Modified by Suchita Bhinge... September 14 2018
    Y = np.zeros_like(X)

    # to store the change in W or cost in each iteration
    W_change = []

    # Main Iteration Loop
    for iteration in range(max_iter):
        term_criterion = 0

        # Current estimated sources
        Y = np.einsum('NnK, nTK -> NTK', W, X)

        # Some additional computations of performance via ISI when True A is supplied
        if supply_A:
            avg_isi, joint_isi = _bss_isi(W, A_w)
            isi[iteration] = joint_isi

        W_old = np.copy(W)  # save current W as W_old
        cost[iteration], _ = _comp_l_sos_cost(W, Y)

        Q = 0
        R = 0

        # Loop over each SCV
        for n in range(N):

            # Efficient version of Sigma_n = 1/T * Y_n @ Y_n.T
            Sigma_n = np.eye(K, dtype=X.dtype)
            for k1 in range(K):
                for k2 in range(k1, K):
                    Sigma_n[k1, k2] = W[n, :, k1] @ R_xx[:, :, k1, k2] @ W[n, :, k2]
                    Sigma_n[k2, k1] = Sigma_n[k1, k2]

            hnk, Q, R = _decouple_trick(W, n, Q, R)

            # Derivative of cost function with respect to wnk
            yn = Y[n, :, :].T

            for k in range(K):
                Sigma_inv = np.linalg.inv(Sigma_n)
                gipyn = np.sum(yn * (Sigma_inv @ yn), axis=0)
                phi = (Sigma_inv[k, :] @ yn) * gipyn ** (
                    -0.5) * gammaShape  # .. Modified by Suchita Bhinge... September 14 2018
                dW = 1 / T * X[:, :, k] @ phi - hnk[:, k] / (W[n, :, k] @ hnk[:, k])
                if grad_projection:  # non-colinear direction normalized
                    dW = _normalize_column_vectors(dW - (W[n, :, k] @ dW) * W[n, :, k])
                W[n, :, k] = _normalize_column_vectors(W[n, :, k] - alpha0 * dW)

                yn[k, :] = W[n, :, k] @ X[:, :, k]
                for kk in range(K):
                    Sigma_n[k, kk] = W[n, :, k] @ R_xx[:, :, k, kk] @ W[n, :, kk]
                Sigma_n[:, k] = Sigma_n[k, :].T

        # Calculate termination criterion
        if termination_criterion == 'change_in_W':
            for k in range(K):
                term_criterion = np.maximum(term_criterion, np.amax(
                    1 - np.abs(np.diag(W_old[:, :, k] @ W[:, :, k].T))))

        elif termination_criterion == 'change_in_cost':
            if iteration == 0:
                term_criterion = 1
            else:
                term_criterion = np.abs(cost[iteration - 1] - cost[iteration]) / np.abs(
                    cost[iteration])

        W_change.append(term_criterion)

        # Decrease step size alpha if cost increased from last iteration
        if iteration > 0 and cost[iteration] > cost[iteration - 1]:
            alpha0 = np.maximum(alpha_min, alpha_scale * alpha0)

        # Check the termination condition
        if term_criterion < W_diff_stop or iteration == max_iter - 1:
            break
        elif term_criterion > blowup or np.isnan(cost[iteration]):
            for k in range(K):
                W[:, :, k] = np.eye(N) + 0.1 * np.random.randn(N, N)
            if X.dtype == complex:
                W = W + 0.1j * np.random.randn(N, N, K)
            if verbose:
                print('W blowup, restart with new initial value.')

        # Display Iteration Information
        if verbose:
            if supply_A:
                print(f'Step {iteration}: W change: {term_criterion}, Cost: {cost[iteration]}, '
                      f'Avg ISI: {avg_isi}, Joint ISI: {joint_isi}')
            else:
                print(f'Step {iteration}: W change: {term_criterion}, Cost: {cost[iteration]}')

    # Finish Display
    if iteration == 0 and verbose:
        if supply_A:
            print(
                f'Step {iteration}: W change: {term_criterion}, Cost: {cost[iteration]}, '
                f'Avg ISI: {avg_isi}, Joint ISI: {joint_isi}')
        else:
            print(f'Step {iteration}: W change: {term_criterion}, Cost: {cost[iteration]}')

    # Clean-up Outputs
    cost = cost[0:iteration + 1]

    if supply_A:
        isi = isi[0:iteration + 1]

    if whiten:
        for k in range(K):
            W[:, :, k] = W[:, :, k] @ V[:, :, k]
    else:  # no prewhitening
        # Scale demixing vectors to generate unit variance sources
        for n in range(N):
            for k in range(K):
                W[n, :, k] /= np.sqrt(W[n, :, k] @ R_xx[:, :, k, k] @ W[n, :, k])

    # Resort order of SCVs: Order the components from most to least ill-conditioned
    if not whiten:
        V = None
    W, Sigma_N = _resort_scvs(W, R_xx, whiten, V)

    end = time.time()
    print(f"IVA-L-SOS finished after {(end - start) / 60} minutes with {iteration} iterations.")

    if return_W_change:
        return W, cost, Sigma_N, isi, W_change
    else:
        return W, cost, Sigma_N, isi



#     Copyright (c) <2021> <University of Paderborn>
#     Signal and System Theory Group, Univ. of Paderborn, https://sst-group.org/
#     https://github.com/SSTGroup/independent_vector_analysis
#
#     Permission is hereby granted, free of charge, to any person
#     obtaining a copy of this software and associated documentation
#     files (the "Software"), to deal in the Software without restriction,
#     including without limitation the rights to use, copy, modify and
#     merge the Software, subject to the following conditions:
#
#     1.) The Software is used for non-commercial research and
#        education purposes.
#
#     2.) The above copyright notice and this permission notice shall be
#        included in all copies or substantial portions of the Software.
#
#     3.) Publication, Distribution, Sublicensing, and/or Selling of
#        copies or parts of the Software requires special agreements
#        with the University of Paderborn and is in general not permitted.
#
#     4.) Modifications or contributions to the software must be
#        published under this license. The University of Paderborn
#        is granted the non-exclusive right to publish modifications
#        or contributions in future versions of the Software free of charge.
#
#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#     EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#     OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#     NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#     HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#     WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#     FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#     OTHER DEALINGS IN THE SOFTWARE.
#
#     Persons using the Software are encouraged to notify the
#     Signal and System Theory Group at the University of Paderborn
#     about bugs. Please reference the Software in your publications
#     if it was used for them.


import numpy as np
import scipy as sc
import scipy.linalg


def _jbss_sos(x, tau=0, cost_selection='cross'):
    """
    Joint blind source separation using second order statistics.


    Parameters
    ----------
    x : np.ndarray
        the mixtures of dimensions N x T x K

    tau : int or np.ndarray
        the non-negative time delays, e.g. 0 or np.arange(5) or np.array([0,3,7,12]).
        Always assume tau[0]=0

    cost_selection : str, optional
        'cross' or 'whole'. In case of 'cross', only the cross cost is minimized by using
        sequential orthogonal procrustes. In case of 'whole', a gradient search algorithm is used,
        starting from the initial guess provided by sequential orthogonal procrustes.


    Returns
    -------
    W : np.ndarray
        demixing matrix of dimensions N x N x K


    Notes
    -----
    For a general description of the algorithm and its relationship with others,
    see http://mlsp.umbc.edu/jointBSS_introduction.html

    References:
    [1] X.-L. Li, T. Adali, & M. Anderson, "Joint Blind Source Separation by Generalized Joint
    Diagonalization of Cumulant Matrices," Signal Process., 2011, 91, 2314-2322
    [2] X.-L. Li, M. Anderson, & T. Adali, "Second and Higher-Order Correlation Analysis of
    Multiset Multidimensional Variables by Joint Diagonalization," Lecture Notes in Computer
    Science: Independent Component Analysis and Blind Signal Separation, Latent Variable Analysis
    and Signal Separation, Springer Berlin / Heidelberg, 2010, 6365, 197-204

    Code is converted to Python by Isabell Lehmann (isabell.lehmann at sst.upb.de)

    """

    N, T, K = x.shape

    if type(tau) is not int:
        tau_ = np.copy(tau)
        tau_[0] = 0
    else:
        tau_ = np.expand_dims(tau, axis=0)
    tau_ = np.abs(tau_)
    L = tau_.size

    # first remove mean
    x_w = np.zeros_like(x)
    for k in range(K):
        for n in range(N):
            x_w[n, :, k] = x[n, :, k] - np.mean(x[n, :, k])

    # second whitening the mixtures
    P = np.zeros((N, N, K), dtype=x.dtype)
    for k in range(K):
        R = 1 / T * x_w[:, :, k] @ np.conj(x_w[:, :, k].T)
        P[:, :, k] = np.linalg.inv(sc.linalg.sqrtm(R))
        x_w[:, :, k] = P[:, :, k] @ x_w[:, :, k]

    # third construct the correlation matrix at L time-delays
    C = np.zeros((N, N, K, K, L), dtype=x.dtype)
    for k1 in range(K):
        for k2 in range(K):
            for el in range(L):
                C[:, :, k1, k2, el] = 1 / T * x_w[:, 0:T - tau_[el], k1] @ np.conj(
                    x_w[:, tau_[el]:T, k2].T)

    # joint diagonalization
    W0 = np.zeros((N, N, K), dtype=x.dtype)
    for k in range(K):
        W0[:, :, k] = np.eye(N)

    if cost_selection == 'whole':
        W0 = _diag_cross(C, W0, 1e-8)
        W = _diag_whole(C, W0)

    elif cost_selection == 'cross':
        if L == 1:
            W = _sym_diag_cross(C, W0, 1e-8)
        else:
            W = _diag_cross(C, W0, 1e-8)

    else:
        raise ValueError("cost_selection must be 'whole' or 'cross'")

    # recovery the whitening
    for k in range(K):
        W[:, :, k] = W[:, :, k] @ P[:, :, k]

    return W


def _diag_cross(T, W0=None, tol=1e-10):
    """
    Joint diagonalization algorithm minimizing the cross cost by sequential orthogonal procrustes.


    Parameters
    ----------
    T : np.ndarray
        target matrices of dimensions N x N x K x K x L

    W0 : np.ndarray, optional
        initial guess of demixing matrix of dimensions N x N x K

    tol : float, optional
        stopping condition


    Returns
    -------
    W : np.ndarray
        demixing matrix of dimensions N x N x K

    Code is converted to Python by Isabell Lehmann (isabell.lehmann at sst.upb.de)

    """

    N, _, K, _, L = T.shape
    D = np.zeros((N, K, K, L), dtype=T.dtype)

    if W0 is None:
        W0 = np.zeros((N, N, K))
        for k in range(K):
            W0[:, :, k] = np.random.randn(N, N)
            W0[:, :, k] = np.linalg.solve(sc.linalg.sqrtm(W0[:, :, k] @ W0[:, :, k].T), W0[:, :, k])
    W = np.copy(W0)

    A = np.zeros((2 * L * N * (K - 1), N, K), dtype=T.dtype)
    for p in range(K):
        cnt = 0
        for q in range(K):
            if q != p:
                for el in range(L):
                    A[cnt * N:(cnt + 1) * N, :, p] = np.conj(T[:, :, p, q, el].T)
                    cnt = cnt + 1
                    A[cnt * N:(cnt + 1) * N, :, p] = T[:, :, q, p, el]
                    cnt = cnt + 1

    max_iter = 1000
    cost = np.zeros(max_iter)
    last_W = np.copy(W)

    for iter in range(max_iter):

        cost[iter] = 0
        for p in range(K):
            W_p = W[:, :, p]
            for q in range(K):
                if q != p:
                    W_qh = np.conj(W[:, :, q].T)
                    for el in range(L):
                        D[:, p, q, el] = np.diag(W_p @ T[:, :, p, q, el] @ W_qh)

        B = np.zeros((2 * L * N * (K - 1), N), dtype=T.dtype)
        randp = np.random.permutation(K)
        for p0 in range(K):
            p = randp[p0]

            cnt = 0
            for q in range(K):
                if q != p:
                    for el in range(L):
                        B[cnt * N:(cnt + 1) * N, :] = np.conj(W[:, :, q].T) @ np.diag(
                            np.conj(D[:, p, q, el]))
                        cnt += 1
                        B[cnt * N: (cnt + 1) * N, :] = np.conj(W[:, :, q].T) @ np.diag(
                            D[:, q, p, el])
                        cnt += 1
            U, _, Vh = np.linalg.svd(np.conj(B.T) @ A[:, :, p])
            W[:, :, p] = U @ Vh

        tol_ = 0
        for k in range(K):
            tol_ = np.maximum(tol_, np.amax(1 - np.diag(W[:, :, k] @ np.conj(last_W[:, :, k].T))))
        if tol_ < tol:
            break
        else:
            last_W = np.copy(W)

    return W


def _diag_whole(T, W0, tol=1e-5):
    """
    Joint diagonalization algorithm minimizing the whole cost


    Parameters
    ----------
    T : np.ndarray
        target matrices of dimensions N x N x K x K x L

    W0 : np.ndarray
        initial guess of demixing matrix of dimensions N x N x K

    tol : float, optional
        stopping condition


    Returns
    -------
    W : np.ndarray
        demixing matrix of dimensions N x N x K

    Code is converted to Python by Isabell Lehmann (isabell.lehmann at sst.upb.de)

    """

    W = np.copy(W0)
    mu = 1 / 20
    W = _iva_diag_kernel(T, W, mu, tol)
    return W


def _iva_diag_kernel(T, W0, mu, tol=1e-5):
    """
    The actual gradient search algorithm

    Parameters
    ----------
    T : np.ndarray
        target matrices of dimensions N x N x K x K x L

    W0 : np.ndarray
        initial guess of demixing matrix of dimensions N x N x K

    mu : float
        step size

    tol : float, optional
        stopping condition


    Returns
    -------
    W : np.ndarray
        demixing matrix of dimensions N x N x K

    Code is converted to Python by Isabell Lehmann (isabell.lehmann at sst.upb.de)

    """

    N, _, K, _, L = T.shape

    max_iter = 1000
    cost = np.zeros(max_iter)
    min_cost = np.inf
    best_W = np.copy(W0)
    last_W = np.copy(W0)
    W = np.copy(W0)
    cost_increase_counter = 0
    # tol = 1
    for iter in range(max_iter):

        cost[iter] = 0
        for p in range(K):
            for q in range(K):
                for el in range(L):
                    B = W[:, :, p] @ T[:, :, p, q, el] @ np.conj(W[:, :, q].T)
                    B = B - np.diag(np.diag(B))
                    cost[iter] += np.linalg.norm(B) ** 2

        grad = np.zeros((N, N, K), dtype=T.dtype)
        for p in range(K):
            for el in range(L):
                for q in range(K):
                    B = W[:, :, p] @ T[:, :, p, q, el] @ np.conj(W[:, :, q].T)
                    B = B - np.diag(np.diag(B))
                    grad[:, :, p] = grad[:, :, p] + B @ W[:, :, q] @ np.conj(T[:, :, p, q, el].T)

                    B = W[:, :, p] @ np.conj(T[:, :, q, p, el].T) @ np.conj(W[:, :, q].T)
                    B = B - np.diag(np.diag(B))
                    grad[:, :, p] = grad[:, :, p] + B @ W[:, :, q] @ T[:, :, q, p, el]

            grad[:, :, p] -= W[:, :, p] @ np.conj(grad[:, :, p].T) @ W[:, :, p]
            grad[:, :, p] /= np.linalg.norm(grad[:, :, p], ord=2)  # normalization

            W[:, :, p] = W[:, :, p] - mu * grad[:, :, p]
            if W.dtype == complex:
                W[:, :, p] = np.linalg.solve(sc.linalg.sqrtm(W[:, :, p] @ np.conj(W[:, :, p].T)),
                                             W[:, :, p])
            else:  # use real part of sqrtm since imaginary is 0
                W[:, :, p] = np.linalg.solve(np.real(sc.linalg.sqrtm(W[:, :, p] @ W[:, :, p].T)),
                                             W[:, :, p])

        if cost[iter] < min_cost:
            min_cost = cost[iter]
            best_W = np.copy(last_W)

        if cost[iter] > min_cost:
            cost_increase_counter += 1
            if cost_increase_counter > 1:
                cost_increase_counter = 0
                mu = 0.5 * mu
                W = np.copy(best_W)
                last_W = np.copy(W)
                continue  # skip code below and jump to next iteration

        tol_ = 0
        for k in range(K):
            tol_ = np.maximum(tol_, np.amax(1 - np.diag(W[:, :, k] @ np.conj(last_W[:, :, k].T))))
        if tol_ < tol:
            break
        else:
            last_W = np.copy(W)

    W = np.copy(best_W)

    return W


def _sym_diag_cross(T, W0=None, tol=1e-10):
    """
    Joint diagonalization algorithm minimizing the cross cost by sequential orthogonal procrustes.
    Assume that T is symmetric, i.e., T[:,:,k1,k2,:] = T[:,:,k2,k1,:].T, and only operate on
    T[:,:,k1,k2,:], k2 > k1


    Parameters
    ----------
    T : np.ndarray
        target matrices of dimensions N x N x K x K x L

    W0 : np.ndarray
        initial guess of demixing matrix of dimensions N x N x K

    tol : float, optional
        stopping condition


    Returns
    -------
    W : np.ndarray
        demixing matrix of dimensions N x N x K

    Code is converted to Python by Isabell Lehmann (isabell.lehmann at sst.upb.de)

    """

    N, _, K, _, L = T.shape
    D = np.zeros((N, K, K, L), dtype=T.dtype)

    if W0 is None:
        W0 = np.zeros((N, N, K), dtype=T.dtype)
        for k in range(K):
            if T.dtype == complex:
                W0[:, :, k] = np.random.randn(N, N) + 1j * np.random.randn(N, N)
            else:
                W0[:, :, k] = np.random.randn(N, N)

            W0[:, :, k] = np.linalg.solve(sc.linalg.sqrtm(W0[:, :, k] * W0[:, :, k].T), W0[:, :, k])

    W = np.copy(W0)

    max_iter = 1000
    cost = np.zeros(max_iter)
    last_W = np.copy(W)

    for iter in range(max_iter):
        cost[iter] = 0
        for p in range(K):
            W_p = W[:, :, p]
            for q in range(p + 1, K):
                W_qh = np.conj(W[:, :, q].T)
                for el in range(L):
                    D[:, p, q, el] = np.diag(W_p @ T[:, :, p, q, el] @ W_qh)

        A = np.zeros((L * N * (K - 1), N), dtype=T.dtype)
        B = np.zeros((L * N * (K - 1), N), dtype=T.dtype)
        randp = np.random.permutation(K)
        for p0 in range(K):
            p = randp[p0]

            cnt = 0
            for q in range(p + 1, K):
                for el in range(L):
                    A[cnt * N:(cnt + 1) * N, :] = np.conj(T[:, :, p, q, el].T)
                    B[cnt * N:(cnt + 1) * N, :] = np.conj(W[:, :, q].T) @ np.diag(
                        np.conj(D[:, p, q, el].T))
                    cnt = cnt + 1
            for q in range(p):
                for el in range(L):
                    A[cnt * N:(cnt + 1) * N, :] = T[:, :, q, p, el]
                    B[cnt * N:(cnt + 1) * N, :] = np.conj(W[:, :, q].T) @ np.diag(D[:, q, p, el])
                    cnt = cnt + 1

            U, _, Vh = np.linalg.svd(np.conj(B.T) @ A)
            W[:, :, p] = U @ Vh

        tol_ = 0
        for k in range(K):
            tol_ = np.maximum(tol_, np.amax(1 - np.diag(W[:, :, k] @ np.conj(last_W[:, :, k].T))))
        if tol_ < tol:
            break
        else:
            last_W = np.copy(W)

    return W


def _cca(X):
    """
    Performs canonical correlation analysis.


    Parameters
    ----------
    X : np.ndarray
        data of two datasets, with dimensions N x T x 2


    Returns
    -------
    W : np.ndarray
        transformation matrices for the two datasets, with dimensions N x N x 2

    Code is converted to Python by Isabell Lehmann (isabell.lehmann at sst.upb.de)

    """

    N, T, K = X.shape

    if K != 2:
        raise ValueError('CCA only works for two datasets')

    R_xx = np.zeros((N, N, 2, 2), dtype=X.dtype)
    for k1 in range(2):
        for k2 in range(2):
            R_xx[:, :, k1, k2] = 1 / T * X[:, :, k1] @ X[:, :, k2].T

    temp = np.linalg.solve(R_xx[:, :, 1, 1], R_xx[:, :, 1, 0])
    A = R_xx[:, :, 0, 1] @ temp
    _, W0 = sc.linalg.eig(A, R_xx[:, :, 0, 0])
    W1 = temp @ W0
    W = np.zeros((N, N, 2), dtype=X.dtype)
    W[:, :, 0] = W0.T
    W[:, :, 1] = W1.T

    return W



#     Copyright (c) <2021> <University of Paderborn>
#     Signal and System Theory Group, Univ. of Paderborn, https://sst-group.org/
#     https://github.com/SSTGroup/independent_vector_analysis
#
#     Permission is hereby granted, free of charge, to any person
#     obtaining a copy of this software and associated documentation
#     files (the "Software"), to deal in the Software without restriction,
#     including without limitation the rights to use, copy, modify and
#     merge the Software, subject to the following conditions:
#
#     1.) The Software is used for non-commercial research and
#        education purposes.
#
#     2.) The above copyright notice and this permission notice shall be
#        included in all copies or substantial portions of the Software.
#
#     3.) Publication, Distribution, Sublicensing, and/or Selling of
#        copies or parts of the Software requires special agreements
#        with the University of Paderborn and is in general not permitted.
#
#     4.) Modifications or contributions to the software must be
#        published under this license. The University of Paderborn
#        is granted the non-exclusive right to publish modifications
#        or contributions in future versions of the Software free of charge.
#
#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#     EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#     OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#     NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#     HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#     WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#     FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#     OTHER DEALINGS IN THE SOFTWARE.
#
#     Persons using the Software are encouraged to notify the
#     Signal and System Theory Group at the University of Paderborn
#     about bugs. Please reference the Software in your publications
#     if it was used for them.


import numpy as np
import scipy.linalg
from scipy.special import gammaln


def normalize_row_vectors(x):
    return x / np.linalg.norm(x, axis=1)[:, np.newaxis]


def generate_sources(rho, N, T, K):
    """
    Generate sources with simple correlation structure:
    For i = 1, ..., N, the ith source in each of the K datasets is correlated.

    Parameters
    ----------

    rho : float or np.ndarray
        correlation value or array of dimension N containing the correlation value for each SCV

    N : int
        number of sources in one source vector

    T : int
        number of samples in each source

    K : int
        number of datasets

    Returns
    -------
    S : np.ndarray
        matrix of dimensions N x T x K, with S[n,:,k] is the nth source in the kth dataset

    """

    # fit shape of rho
    if type(rho) == float:
        rho = np.ones(N) * rho
    else:
        assert (N == rho.shape[0]), 'rho has incorrect shape, must have shape N'

    # joint mean
    mu = np.zeros(N * K)

    # joint covariance matrix
    R = np.eye(K * N)
    for i in range(K):
        for j in range(i, K):
            # cross covariance matrices
            if i != j:
                R[i * N:(i + 1) * N, j * N:(j + 1) * N] = np.diag(rho)
                R[j * N:(j + 1) * N, i * N:(i + 1) * N] = R[i * N:(i + 1) * N, j * N:(j + 1) * N]

    # draw source vector of joint distribution
    all_sources = np.random.multivariate_normal(mu, R, T).T

    S_matrix = np.zeros((N, T, K))
    for i in range(N * K):
        n = np.mod(i, N)
        k = i // N
        S_matrix[n, :, k] = all_sources[i, :]

    return S_matrix


def MGGD_generation(N, p=None, correlation_structure=None, rho=None, beta=1, Sigma=None):
    """
    Contains five functions that can be used to generate multivariate generalized Gaussian (MGGD)
    distributed sources. If Sigma is given, none of the functions will be used but data will be
    generated using Sigma.
    
    
    Parameters
    ----------
    N : int
        Number of realizations / samples
        
    p : int, optional
        Dimension / number of variables. Must be at least 2. Must be given if Sigma is None.

    correlation_structure : str, optional
        Which structure of correlation matrix to use. Must be given if Sigma is None.
            - 'uniform': elements are equal to rho
            - 'ar': cov[i, j] = rho ** (np.abs(i - j))
            - 'q_qt': normalized random samples ~ U[-0.2, 0.8) multiplied by themselves
            - 'two_rho': most elements are equal, specific elements have another value
            - 'rho_list': elements of each row/column of the covariance matrix have a specific value
            - 'block': block-diagonal matrix. Elements in a block have same value, non
                       block-elements have same value, diagonal elements are 1
        
    rho : float or dict or np.ndarray or list, optional
        Correlation value. Must be given if Sigma is None and correlation_structure is not 'q_qt'.
        If correlation_structure=='two_rho', rho must be a dictionary with keys 'val' and 'idx'.
            'val' is a np.ndarray or list with two values of correlation: 'val[0]' provides the
            basic correlation.
            'idx' is a np.ndarray or list that gives the specific locations where 'val[1]' is set.
        If correlation_structure=='rho_list', rho must be a np.ndarray or list with length p-1.
        If correlation_structure=='block', rho is a dict with keys 'blocks' and 'val'.
            - rho['blocks'] must be tuple or list of tuples, where each tuple contains
              (correlation value of the block, start idx of the block, length of the blocks).
            - rho['val'] is the correlation value of the non-block elements.
        
    beta : float, optional
        Shape parameter.
            - If beta is equal to one, then we generate Gaussian sources.
            - If beta is less than 1, the distribution of the marginals is more peaky, with heavier
              tails.
            - If beta is greater than 1, it is less peaky with lighter tails.

    Sigma : np.ndarray, optional
        User-defined covariance matrix of dimensions p x p, where p is the number of sources

    Returns
    -------
    X : np.ndarray
        matrix of dimensions p x N, which contains N samples of each of the p MGGD sources

    COV: np.ndarray
        covariance matrix
        
    
    Notes
    -----
    Coded by Lionel Bombrun (lionel.bombrun at u-bordeaux.fr)
    and Zois Boukouvalas (zb1 at umbc.edu)
    Modified by Q. Long 08/2018 (qunfang1 at umbc.edu)
    Converted to Python by Isabell Lehmann (isabell.lehmann at sst.upb.de)

    Machine Learning for Signal Processing Lab (MLSP-Lab)
    http://mlsp.umbc.edu

    Reference: E. Gomez, M. Gomez-Villegas, and J. Marin, "A multivariate generalization of the
    power exponential family of distributions," Communications in Statistics-Theory and Methods,
    vol. 27, no. 3, pp. 589{600}, 1998.

    """

    if type(N) is not int or N < 2:
        raise AssertionError("'N' must be an integer and must be at least 2.")

    if Sigma is None:

        if p is None:
            raise AssertionError("Either 'p' or 'Sigma' must be given.")
        elif correlation_structure is None:
            raise AssertionError("Either 'correlation_structure' or 'Sigma' must be given.")
        elif rho is None and correlation_structure != 'q_qt':
            raise AssertionError("Either 'rho' or 'Sigma' must be given.")

        if type(p) is not int or p < 2:
            raise AssertionError("'p' must be an integer and must be at least 2.")

        if correlation_structure == 'two_rho':
            if type(rho) is not dict:
                raise AssertionError("'rho must be dict with rho['val'] = np.ndarray, "
                                     "rho['idx'] = np.ndarray.")
            else:
                if sorted(list(rho.keys())) != ['idx', 'val']:
                    raise AssertionError("'rho' must have keys 'val' and 'idx'.")
                elif (type(rho['val']) is not np.ndarray and type(rho['val']) is not list) or (
                        type(rho['idx']) is not np.ndarray and type(rho['idx']) is not list):
                    raise AssertionError("'rho[key]' must be np.ndarray or list.")
                elif len(rho['val']) != 2:
                    raise AssertionError("'rho[val]' must consists of 2 elements.")

        elif correlation_structure == 'rho_list':
            if type(rho) is not list and type(rho) is not np.ndarray:
                raise AssertionError("'rho' must be np.ndarray or list.")

        elif correlation_structure == 'block':
            if type(rho) is not dict:
                raise AssertionError("'rho must be dict with rho['val'] = np.ndarray, "
                                     "rho['blocks'] = tuple or list of tuples.")
            else:
                if sorted(list(rho.keys())) != ['idx', 'blocks']:
                    raise AssertionError("'rho' must have keys 'val' and 'blocks'.")
                elif type(rho['val']) is not float:
                    raise AssertionError("'rho[val]' must be a float value.")
                if type(rho['blocks']) is not list and type(rho['block']) is not tuple:
                    raise AssertionError("'rho[blocks]' must be tuple or list of tuples.")
        else:
            if type(rho) is not float and type(rho) is not int:
                raise AssertionError("'rho' must be a float or int value.")

        if correlation_structure not in ['uniform', 'ar', 'two_rho', 'rho_list', 'q_qt']:
            raise AssertionError("'correlation_structure' must be 'uniform', 'ar', 'two_rho', "
                                 "'rho_list' or 'q_qt'.")

    else:
        p = Sigma.shape[0]

        if Sigma.ndim != 2:
            raise AssertionError("'Sigma' must be a 2D matrix.")
        elif Sigma.shape[0] != Sigma.shape[1]:
            raise AssertionError("Sigma' must be a square matrix.")

    # ---------Q. Long 02/2018--------########
    betah = 2 * beta
    etanln = 1 / beta * np.log(2) + gammaln((p + 2) / betah) - gammaln(p / betah) - np.log(p)
    etan = np.exp(etanln)
    # ---------Q. Long 02/2018--------########

    X = _random_sphere(N, p).T

    if Sigma is None:
        # This is the covariance matrix we want
        COV1 = np.abs(_create_covariance_matrix(p, correlation_structure, rho))
    else:
        COV1 = np.copy(Sigma)

    COV = COV1 / etan  # Use this as the Sigma in Gomez's paper

    # check if covariance matrix is positive semidefintive
    try:
        # Use Cholesky decomposition instead of the square root of Sigma
        X = np.linalg.cholesky(COV) @ X
    except:
        raise AssertionError("Covariance matrix is not positive semidefinite.")

    tau = np.random.gamma(p / (2 * beta), scale=2, size=N) ** (1 / (2 * beta))

    X = tau * X

    return X, COV


def _create_covariance_matrix(p, correlation_structure, rho):
    if correlation_structure == 'uniform':
        Sigma = np.full((p, p), rho)
        np.fill_diagonal(Sigma, 1)

    elif correlation_structure == 'ar':
        Sigma = np.eye(p)
        for i in range(p):
            for j in range(p):
                Sigma[i, j] = rho ** (np.abs(i - j))

    elif correlation_structure == 'q_qt':
        R = normalize_row_vectors(np.random.rand(p, p) - 0.2)
        Sigma = R @ R.T

    elif correlation_structure == 'two_rho':
        Sigma = np.full((p, p), rho['val'][0])
        for i in rho['idx']:
            for j in rho['idx']:
                Sigma[i, j] = rho['val'][1]
        np.fill_diagonal(Sigma, 1)

    elif correlation_structure == 'rho_list':
        Sigma = np.eye(p)
        n = len(rho)
        if np.mod(p - 1, n):
            raise AssertionError("'rho' must contain 'p'-1 elements")
        M = (p - 1) // n
        for j in range(n):
            Sigma[0, M * j + 1:M * (j + 1) + 1] = rho[j] * np.ones(M)
            Sigma[M * j + 1:M * (j + 1) + 1, 0] = rho[j] * np.ones(M)
        for i in range(1, p):
            for j in range(i + 1, p):
                Sigma[i, j] = Sigma[i - 1, j]
                Sigma[j, i] = Sigma[i, j]

    elif correlation_structure == 'block':
        if type(rho['blocks']) is tuple:
            csl = [rho['blocks']]
        else:
            csl = rho['blocks']

        Sigma = np.full((p, p), rho['val'])
        for (correlation, start, length) in csl:
            Sigma[start:start + length, start:start + length] = correlation
        np.fill_diagonal(Sigma, 1)

    return Sigma


def _random_sphere(N, dim=3):
    """
    Generates uniform random points on the surface of a unit radius N-dimensional sphere centered
    in the origin. This script uses different algorithms according to the dimensions of points:
       -2D:  random generation of theta [0 2*pi]
       -3D:  the "trig method".
       -nD:  Gaussian distribution


    Parameters
    ----------
    N : int
        Number of points to be generated

    dim : int, optional
        Dimension of data: must be at least 2


    Returns
    -------
    X : np.ndarray
        Matrix of dimensions N x dim representing the coordinates of random points generated


    Notes
    -----
    Authors: Luigi Giaccari, Ed Hoyle
    Converted to Python by Isabell Lehmann

    """

    if dim == 2:
        # just a random generation of theta
        temp = np.random.rand(N) * 2 * np.pi  # theta: use as temp value
        X = np.zeros((N, dim))
        X[:, 0] = np.cos(temp)  # x
        X[:, 1] = np.sin(temp)  # y

    elif dim == 3:
        # trig method
        X = np.zeros((N, dim))
        X[:, 2] = np.random.rand(N) * 2 - 1  # z
        t = np.random.rand(N) * 2 * np.pi
        r = np.sqrt(1 - X[:, 2] ** 2)
        X[:, 0] = r * np.cos(t)  # x
        X[:, 1] = r * np.sin(t)  # y

    else:
        # use gaussian distribution
        X = np.random.randn(N, dim)
        X /= np.sqrt(np.sum(X ** 2, axis=1))[:, np.newaxis]

    return X


def randmv_laplace(d, T, lambda_=1, mu=None, gamma=None):
    """

    Generate T iid samples of the d-dimensional multivariate Laplace (ML) distribution, as given in
    "On the Multivariate Laplace Distribution", by Eltoft in IEEE Sig. Proc. Letters 2006.

    Parameters
    ----------
    d : int
        dimension of the multivariate Laplacian

    T : int
        number of iid samples

    lambda_ : float
        exponential rate parameter, > 0

    mu : np.ndarray
        mean vector of dimension d

    gamma :
        internal covariance structure, note det(Gamma)=1


    Returns
    -------
    Y : np.ndarray
        ML source matrix of dimension d x_sorted T


    Notes
    -----

    Note that a method for transforming an uncorrelated ML, Y, into a correlated ML, V, is given by
    Eltoft reference using V=A@Y+b, where A is a d x d real-valued matrix and b is d-dimensional
    real-valued vector, then V is ML with parameters specified in equations (14)-(16):

    lambda_new = lambda_ * np.abs(np.linalg.det(A))**(1/d)
    mu_new = A @ mu + b
    Gamma_new = A @ Gamma @ np.conj(A.T) * np.abs(np.linalg.det(A))**(-2/d)

    The special case of mu=b=0, Gamma=eye(d), and np.linalg.det(A)=1 is nice since
    then, lambda_new=lambda, mu_new=0, and Gamma_new=A @ Gamma @ np.conj(A.T).

    Coded by Matthew Anderson 9/15/2011.
    Converted to Python by Isabell Lehmann 2021-02-01

    """

    if mu is None:
        mu = np.zeros(d)
    if gamma is None:
        gamma = np.eye(d)

    if lambda_ < 0 or np.any(np.iscomplex(lambda_)):
        raise AssertionError('Rate parameter lambda_ should be real-valued and greater than 0.')

    if mu.shape[0] != d or np.any(np.iscomplex(lambda_)):
        raise AssertionError('Mean vector should be real-valued and of dimension d.')

    if gamma.shape[0] != d or gamma.shape[1] != d or np.abs(
            np.linalg.det(gamma) - 1) > 0.0001 or np.any(np.iscomplex(gamma)):
        raise AssertionError(
            'Internal covariance structure needs to be a real-valued square matrix with a determinant of one.')

    X = np.random.randn(d, T)
    Z = np.sqrt(np.random.exponential(1 / lambda_, T))
    Y = mu[:, np.newaxis] + Z * (scipy.linalg.sqrtm(gamma) @ X)  # equation (6)

    return Y



from setuptools import setup


def readme():
    with open('README.md') as f:
        return f.read()


setup(name='independent_vector_analysis',
      version='0.2.1',
      description='Implementation of IVA-G and IVA-L-SOS',
      long_description=readme(),
      long_description_content_type='text/markdown',
      classifiers=[
          'Programming Language :: Python :: 3',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
      ],
      keywords='iva, independent vector analysis, bss, multiset analysis',
      url='https://github.com/SSTGroup/independent_vector_analysis',
      author='Isabell Lehmann',
      author_email='isabell.lehmann@sst.upb.de',
      license='LICENSE',
      packages=['independent_vector_analysis'],
      python_requires='>=3.6',
      install_requires=[
          'numpy',
          'scipy',
          'pytest',
          'joblib',
          'tqdm',
      ],
      include_package_data=True,  # to include non .py-files listed in MANIFEST.in
      zip_safe=False)



#     Copyright (c) <2021> <University of Paderborn>
#     Signal and System Theory Group, Univ. of Paderborn, https://sst-group.org/
#     https://github.com/SSTGroup/independent_vector_analysis
#
#     Permission is hereby granted, free of charge, to any person
#     obtaining a copy of this software and associated documentation
#     files (the "Software"), to deal in the Software without restriction,
#     including without limitation the rights to use, copy, modify and
#     merge the Software, subject to the following conditions:
#
#     1.) The Software is used for non-commercial research and
#        education purposes.
#
#     2.) The above copyright notice and this permission notice shall be
#        included in all copies or substantial portions of the Software.
#
#     3.) Publication, Distribution, Sublicensing, and/or Selling of
#        copies or parts of the Software requires special agreements
#        with the University of Paderborn and is in general not permitted.
#
#     4.) Modifications or contributions to the software must be
#        published under this license. The University of Paderborn
#        is granted the non-exclusive right to publish modifications
#        or contributions in future versions of the Software free of charge.
#
#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#     EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#     OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#     NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#     HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#     WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#     FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#     OTHER DEALINGS IN THE SOFTWARE.
#
#     Persons using the Software are encouraged to notify the
#     Signal and System Theory Group at the University of Paderborn
#     about bugs. Please reference the Software in your publications
#     if it was used for them.


import numpy as np
import scipy as sc
import scipy.linalg
from scipy.special import gamma


def _normalize_column_vectors(x):
    return x / np.linalg.norm(x, axis=0)


def _decouple_trick(W, n, Q=None, R=None):
    """
    Computes the H vector for the decoupling trick [1] of the nth row of W.
    There are many possible methods for computing H. This algorithm just implements the method in
    [1] (in the MATLAB version, there are more methods implemented): A recursive QR algorithm is
    used to compute H.


    Parameters
    ----------
    W : np.ndarray
        stacked demixing matrix of dimensions N x N x K

    n : int
        row index of W for that H vector is calculated

    Q : np.ndarray, optional
        stacked matrix of dimensions N x N x K. Must be provided if n > 0

    R : np.ndarray, optional
        stacked matrix of dimensions N x N-1 x K. Must be provided if n > 0


    Returns
    -------
    H : np.ndarray
        H vector (dimensions N x K) for the decoupling trick [1] of the nth row of W
    Q_ : np.ndarray, optional
        stacked matrix of dimensions N x N x K
    R_ : np.ndarray, optional
        stacked matrix of dimensions N x N-1 x K


    Notes
    -----
    Main References:
    [1] X.-L. Li & X.-D. Zhang, "Nonorthogonal Joint Diagonalization Free of Degenerate Solution,"
    IEEE Trans. Signal Process., 2007, 55, 1803-1814

    Coded by Matthew Anderson (matt.anderson@umbc.edu)
    Converted to Python by Isabell Lehmann (isabell.lehmann@sst.upb.de)

    Version 01 - 20120919 - Initial publication
    Version 02 - 20210129 - Converted to Python

    """

    # get dimensions
    M, N, K = W.shape
    if M != N:
        raise AssertionError('Assuming W is square matrix.')
    H = np.zeros((N, K), dtype=W.dtype)

    # use QR recursive method
    if n == 0:
        Q_ = np.zeros((N, N, K), dtype=W.dtype)
        R_ = np.zeros((N, N - 1, K), dtype=W.dtype)
    else:
        Q_ = np.copy(Q)
        R_ = np.copy(R)

    for k in range(K):
        if n == 0:
            W_tilde = W[1:N, :, k]
            Q_[:, :, k], R_[:, :, k] = np.linalg.qr(np.conj(W_tilde.T), mode='complete')
        else:
            n_last = n - 1
            e_last = np.zeros(N - 1, dtype=W.dtype)
            e_last[n_last] = 1

            # e_last.shape and R.shape[1], W.shape[1] and Q.shape[1] must be equal
            Q_[:, :, k], R_[:, :, k] = sc.linalg.qr_update(Q_[:, :, k], R_[:, :, k],
                                                           -np.conj(W[n, :, k]), e_last)
            Q_[:, :, k], R_[:, :, k] = sc.linalg.qr_update(Q_[:, :, k], R_[:, :, k],
                                                           np.conj(W[n_last, :, k]), e_last)

        H[:, k] = Q_[:, -1, k]

    return H, Q_, R_


def _bss_isi(W, A, s=None):
    """
    Calculate measure of quality of separation for blind source separation algorithms.
    Model:   x = A @ s,   y = W @ x = W @ A @ s


    Parameters
    ----------
    W : np.ndarray
        demixing matrix of dimensions N x N x K or M x p

    A : np.ndarray
        true mixing matrix of dimension N x N x K or p x N

    s : np.ndarray, optional
        true sources of dimension N x T x K or N x T

    (p: #sensors, N: #sources, M: #estimatedsources, K: #datasets, T: #samples)


    Returns
    -------
    avg_isi : float
        avg_isi=0 is optimal.
        Normalized performance index is given in Choi, S. Cichocki, A. Zhang, L. & Amari, S.
        Approximate maximum likelihood source separation using the natural gradient
        Wireless Communications, 2001. (SPAWC '01). 2001 IEEE Third Workshop on Signal
        Processing Advances in, 2001, 235-238.

    joint_isi : float, optional
        joint_isi = 0 is optimal. Only calculated if there are at least 2 datasets, otherwise
        np.nan.
        Normalized joint performance index given in Anderson, Matthew, Tuelay Adali, and
        Xi-Lin Li. "Joint blind source separation with multivariate Gaussian model: Algorithms and
        performance analysis." IEEE Transactions on Signal Processing 60.4 (2011): 1672-1683.

    Notes
    -----
    W is the estimated demixing matrix and A is the true mixing matrix.  It should be noted
    that rows of the mixing matrix should be scaled by the necessary constants such that each
    source has unit variance and accordingly each row of the demixing matrix should be
    scaled such that each estimated source has unit variance.

    Note that A is p x N, where p is the number of sensors and N is the number of signals
    and W is M x p, where M is the number of estimated signals.  Ideally M=N but this is not
    guaranteed.  So if M > N, the algorithm has estimated more sources than it "should", and
    if M < N the algorithm has not found all of the sources.

    Code is converted to Python by Isabell Lehmann (isabell.lehmann at sst.upb.de)
    """

    # generalized permutation invariant flag (default=False), only used when s is None
    gen_perm_inv_flag = False

    if W.ndim == 2 and A.ndim == 2:
        if s is None:
            # Traditional metric, user provided W & A separately
            G = W @ A
            M, N = G.shape
            G_abs = np.abs(G)
            if gen_perm_inv_flag:
                # normalization by row
                G_abs /= np.amax(G_abs, axis=1, keepdims=True)
        else:
            # Equalize energy associated with each estimated source and true source.
            y = W @ A @ s
            # Standard deviation calculated with n-1
            D = np.diag(
                1 / np.std(s, axis=1, ddof=1))  # s_norm = D @ s, where s_norm has unit variance
            U = np.diag(
                1 / np.std(y, axis=1, ddof=1))  # y_norm = U @ y, where y_norm has unit variance

            # Thus: y_norm = U @ W @ A @ np.linalg.inv(D) @ s_norm = G @ s_norm, and
            G = U @ W @ np.linalg.solve(D.T, A.T).T  # A @ np.linalg.inv(D)
            M, N = G.shape
            G_abs = np.abs(G)

        avg_isi = 0
        for m in range(M):
            avg_isi += np.sum(G_abs[m, :]) / np.amax(G_abs[m, :]) - 1

        for n in range(N):
            avg_isi += np.sum(G_abs[:, n]) / np.amax(G_abs[:, n]) - 1

        avg_isi /= (2 * N * (N - 1))

        return avg_isi, np.nan

    elif W.ndim == 3 and A.ndim == 3:
        # IVA/GroupICA/MCCA Metrics
        # For this we want to average over the K groups as well as provide the additional
        # measure of solution to local permutation ambiguity (achieved by averaging the K
        # demixing-mixing matrices and then computing the ISI of this matrix).

        N, M, K = W.shape
        if M != N:
            raise AssertionError('This more general case has not been considered here.')

        avg_isi = 0
        G_abs_total = np.zeros((N, N))
        G = np.zeros((N, N, K), dtype=W.dtype)
        for k in range(K):
            if s is None:
                # Traditional metric, user provided W & A separately
                G_k = W[:, :, k] @ A[:, :, k]
                G_abs = np.abs(G_k)
                if gen_perm_inv_flag:
                    # normalization by row
                    G_abs /= np.amax(G_abs, axis=1, keepdims=True)
            else:

                # Equalize energy associated with each estimated source and true source.
                # Standard deviation calculated with n-1
                y_k = W[:, :, k] @ A[:, :, k] @ s[:, :, k]
                D_k = np.diag(1 / np.std(s[:, :, k], axis=1,
                                         ddof=1))  # s_norm = D @ s, where s_norm has unit variance
                U_k = np.diag(1 / np.std(y_k, axis=1,
                                         ddof=1))  # y_norm = U @ y, where y_norm has unit variance
                # Thus: y_norm = U @ W @ A @ np.linalg.inv(D) @ s_norm = G @ s_norm, and
                G_k = U_k @ W[:, :, k] @ np.linalg.solve(D_k.T,
                                                         A[:, :, k].T).T  # A @ np.linalg.inv(D)
                G_abs = np.abs(G_k)

            G[:, :, k] = G_k

            G_abs_total += G_abs

            for n in range(N):
                avg_isi += np.sum(G_abs[n, :]) / np.amax(G_abs[n, :]) - 1

            for m in range(N):
                avg_isi += np.sum(G_abs[:, m]) / np.amax(G_abs[:, m]) - 1

        avg_isi /= (2 * N * (N - 1) * K)

        G_abs = np.copy(G_abs_total)
        if gen_perm_inv_flag:
            # normalization by row
            G_abs /= np.amax(G_abs, axis=1, keepdims=True)

        joint_isi = 0
        for n in range(N):
            joint_isi += np.sum(G_abs[n, :]) / np.max(G_abs[n, :]) - 1

        for m in range(N):
            joint_isi += np.sum(G_abs[:, m]) / np.max(G_abs[:, m]) - 1

        joint_isi /= (2 * N * (N - 1))

        return avg_isi, joint_isi
    else:
        raise AssertionError('All inputs must be of either dimension 2 or 3')


def whiten_data(x, dim_red=None):
    """
    Whitens the data vector so that E{z z.T} = I, where z = V @ x.
    Optionally, a dimension reduction can be performed.


    Parameters
    ----------
    x : np.ndarray
        data vector of dimensions N x T x K (#sources x #samples x #datasets) or N x T

    dim_red : int, optional
        dimension to which the data should be reduced


    Returns
    -------
    z : np.ndarray
        whitened data of dimension N x T x K or N x T

    V : np.ndarray
        whitening transformation of dimension N x N x K or N x N

    """
    if dim_red is None:
        dim_red = x.shape[0]

    if x.ndim == 2:
        N, T = x.shape

        # Step 1. Center the data.
        x_zm = x - np.mean(x, axis=1, keepdims=True)

        # Step 2. Form MLE of data covariance.
        covar = np.cov(x_zm, ddof=0)

        # Step 3. Eigen decomposition of covariance.
        eigval, eigvec = np.linalg.eigh(covar)

        # sort eigenvalues and corresponding eigenvectors in descending order
        eigval = np.flip(eigval)
        eigvec = np.fliplr(eigvec)

        # Step 4. Forming whitening transformation.
        V = np.einsum('n,Nn -> nN', 1 / np.sqrt(eigval[0:dim_red]), np.conj(eigvec[:, 0:dim_red]))

        # Step 5. Form whitened data
        z = V @ x_zm

    else:
        N, T, K = x.shape

        eigval = np.zeros((N, K), dtype=x.dtype)
        eigvec = np.zeros((N, N, K), dtype=x.dtype)
        x_zm = np.zeros_like(x)

        for k in range(K):
            # Step 1. Center the data.
            x_zm[:, :, k] = x[:, :, k] - np.mean(x[:, :, k], axis=1, keepdims=True)

            # Step 2. Form MLE of data covariance.
            covar = np.cov(x_zm[:, :, k], ddof=0)

            # Step 3. Eigen decomposition of covariance.
            eigval[:, k], eigvec[:, :, k] = np.linalg.eigh(covar)

        # sort eigenvalues and corresponding eigenvectors in descending order
        eigval = np.flipud(eigval)
        eigvec = np.flip(eigvec, 1)

        # efficient implementation of
        #     V[:, :, k] = np.linalg.solve(np.diag(np.sqrt(eigval)), np.conj(eigvec.T))
        #     z[:, :, k] = V[:, :, k] @ x_k

        # Step 4. Forming whitening transformation.
        V = np.einsum('nk,Nnk -> nNk', 1 / np.sqrt(eigval[0:dim_red, :]),
                      np.conj(eigvec[:, 0:dim_red, :]))

        # Step 5. Form whitened data
        z = np.einsum('nNk, Nvk-> nvk', V, x_zm)
    return z, V


def _comp_l_sos_cost(W, Y, const_log=None, scale_sources=False):
    N, T, K = Y.shape
    cost = 0
    xi_shape = K + 1  # .. Modified by Suchita Bhinge... September 14 2018
    if const_log is None:
        const_Ck = 0.5 * gamma(K / 2) * xi_shape ** (K / 2) / (np.pi ** (K / 2) * gamma(K))
        const_log = -np.log(const_Ck)

    if scale_sources:
        for k in range(K):
            ypower = np.diag(1 / T * Y[:, :, k] @ np.conj(Y[:, :, k].T))
            W[:, :, k] /= np.sqrt(ypower)[:, np.newaxis]

    for k in range(K):
        cost -= np.log(np.abs(np.linalg.det(W[:, :, k])))

    for n in range(N):
        yn = Y[n, :, :].T  # K x T
        CovN = 1 / T * yn @ np.conj(yn.T)
        gip = np.sum(np.conj(yn) * np.linalg.solve(CovN, yn), axis=0)
        dcost = (const_log + 0.5 * np.log(np.linalg.det(CovN))) + xi_shape ** 0.5 * np.mean(
            gip ** 0.5)  # .. Modified by Suchita Bhinge... September 14 2018
        cost += dcost
    y = np.copy(Y)

    return cost, y


def _resort_scvs(W, R_xx, whiten=False, V=None, complex_valued=False, circular=False, P_xx=None):
    """
    Resort order of SCVs: Order the components from most to least ill-conditioned
    """

    N, _, K = W.shape

    # First, compute the data covariance matrices (by undoing any whitening)
    if whiten:
        for k1 in range(K):
            for k2 in range(k1, K):
                R_xx[:, :, k1, k2] = np.linalg.solve(V[:, :, k2],
                                                     np.linalg.solve(V[:, :, k1],
                                                                     R_xx[:, :, k1, k2]).T).T
                R_xx[:, :, k2, k1] = np.conj(R_xx[:, :, k1, k2].T)  # R_xx is Hermitian

    # Second, compute the determinant of the SCVs
    if complex_valued:
        detSCV = np.zeros(N, dtype=complex)
        Sigma_N = np.zeros((K, K, N), dtype=complex)
    else:
        detSCV = np.zeros(N)
        Sigma_N = np.zeros((K, K, N))

    for n in range(N):
        # Efficient version of Sigma_n = 1/T * Y_n @ np.conj(Y_n.T) with Y_n = W_n @ X_n
        if complex_valued:
            Sigma_n = np.zeros((K, K), dtype=complex)
        else:
            Sigma_n = np.zeros((K, K))

        for k1 in range(K):
            for k2 in range(k1, K):
                Sigma_n[k1, k2] = W[n, :, k1] @ R_xx[:, :, k1, k2] @ np.conj(W[n, :, k2])
                Sigma_n[k2, k1] = np.conj(Sigma_n[k1, k2])  # sigma_n is Hermitian
        Sigma_N[:, :, n] = Sigma_n

        if complex_valued and not circular:
            Sigma_P_n = np.zeros((K, K), dtype=complex)  # pseudo = 1/T * Y_n @ Y_n.T
            for k1 in range(K):
                for k2 in range(k1, K):
                    Sigma_P_n[k1, k2] = W[n, :, k1] @ P_xx[:, :, k1, k2] @ W[n, :, k2]
                    Sigma_P_n[k2, k1] = Sigma_P_n[k1, k2]  # sigma_P_n is symmetric
            detSCV[n] = np.linalg.det(np.vstack([np.hstack([Sigma_n, Sigma_P_n]),
                                                 np.hstack(
                                                     [np.conj(Sigma_P_n.T), np.conj(Sigma_n)])]))
        else:
            detSCV[n] = np.linalg.det(Sigma_n)

    # Third, sort and apply
    isort = np.argsort(detSCV)
    Sigma_N = Sigma_N[:, :, isort]
    for k in range(K):
        W[:, :, k] = W[isort, :, k]

    return W, Sigma_N
